from enum import Enum


class Environment(Enum):
    PRODUCTION = "prod"
    STAGING = "staging"
    TEST = "test"


API_BASE_URL = "https://eodatahub.org.uk"
S3_BASE_URL_TEMPLATE = (
    "https://{workspace_name}.{environment}.eodatahub-workspaces.org.uk/files/workspaces-eodhp-{environment}"
)
PAGINATION_LIMIT = 10



import os
import time
from collections.abc import Generator

import dotenv
import pytest
import requests


@pytest.fixture(scope="session")
def api_token() -> Generator[str, None, None]:
    dotenv.load_dotenv()
    username = os.getenv("EODH_USERNAME")
    password = os.getenv("EODH_PASSWORD")

    if not password:
        yield "token"
        return

    base_url = "https://staging.eodatahub.org.uk"

    keycloak_realm = os.getenv("EODH_KEYCLOAK_REALM")
    client_id = os.getenv("EODH_CLIENT_ID")

    token_url = f"{base_url}/keycloak/realms/{keycloak_realm}/protocol/openid-connect/token"
    data = {
        "client_id": client_id,
        "username": username,
        "password": password,
        "grant_type": "password",
        "scope": "workspaces",
    }

    response = requests.post(token_url, data=data)
    response.raise_for_status()
    token_data = response.json()
    access_token = token_data.get("access_token")

    token_url = f"{base_url}/api/workspaces/{username}/me/tokens"

    response = requests.post(
        token_url,
        headers={"Authorization": f"Bearer {access_token}"},
        json={"name": f"pyeodh-test-token-{int(time.time())}"},
    )
    response.raise_for_status()

    token = response.json().get("token")
    token_id = response.json().get("id")
    yield token

    token_url = f"{base_url}/api/workspaces/{username}/me/tokens/{token_id}"
    headers = {"Authorization": f"Bearer {access_token}"}

    response = requests.delete(token_url, headers=headers)
    response.raise_for_status()


@pytest.fixture(scope="session")
def username() -> str:
    dotenv.load_dotenv()
    return os.getenv("EODH_USERNAME", "oxidian-test")



from __future__ import annotations

import logging
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from pyeodh import consts
from pyeodh.eodh_object import EodhObject
from pyeodh.types import Headers, Params, RequestMethod

# Avoid circular imports only for type checking
if TYPE_CHECKING:
    from pyeodh.client import Client

T = TypeVar("T", bound=EodhObject)

logger = logging.getLogger(__name__)


class PaginatedList(Generic[T]):
    def __init__(
        self,
        cls: type[T],
        client: Client,
        method: RequestMethod,
        first_url: str,
        list_key: str,
        headers: Headers | None = None,
        params: Params | None = None,
        first_data: dict | None = None,
        parent: EodhObject | None = None,
    ) -> None:
        self._elements: list[T] = []
        self._cls = cls
        self._client = client
        self._method: RequestMethod = method
        self._next_url: str | None = first_url
        self._headers = headers
        self._params = params
        self._total_count: int | None = None
        self._list_key = list_key
        self._data = first_data
        self._parent = parent

    @property
    def total_count(self) -> int | None:
        if not self._total_count:
            data: dict | None = self._data.copy() if self._data is not None else None
            params: Params | None = self._params.copy() if self._params is not None else None
            if data and "limit" in data:
                data["limit"] = 1
            else:
                if params is None:
                    params = {}
                params["limit"] = 1
            if not self._next_url:
                raise RuntimeError("Next url not specified!")
            _, resp_data = self._client._request_json(
                self._method,
                self._next_url,
                headers=self._headers,
                params=params,
                data=data,
            )
            self._total_count = resp_data.get("numMatched")
            if self._total_count is None:
                logger.warning("`numMatched` number is not in the response.")

        return self._total_count

    def __iter__(self) -> Iterator[T]:
        yield from self._elements
        while self._has_next():
            new_elements = self._fetch_next()
            self._elements += new_elements
            yield from new_elements

    def __getitem__(self, index: int | slice) -> Any:
        assert isinstance(index, (int, slice))
        if isinstance(index, int):
            self._fetch_to_index(index)
            return self._elements[index]
        elif isinstance(index, slice):
            return self.PagedSlice(self, index)

    def _fetch_to_index(self, index: int) -> None:
        while len(self._elements) <= index and self._has_next():
            new_elements = self._fetch_next()
            self._elements += new_elements

    def _has_next(self) -> bool:
        return self._next_url is not None

    def _fetch_next(self) -> list[T]:
        if not self._next_url:
            raise RuntimeError("Next url not specified!")
        if "/planet/search?next=" in self._next_url:
            self._next_url, next_token = self._next_url.split("?next=")
            if self._data is None:
                self._data = {}
            self._data["next"] = next_token
        headers, resp_data = self._client._request_json(
            self._method,
            self._next_url,
            headers=self._headers,
            params=self._params,
            data=self._data,
        )
        next_link = next(filter(lambda ln: ln.get("rel") == "next", resp_data.get("links", {})), {})
        self._next_url = next_link.get("href")

        self._data = next_link.get("body")
        self._total_count = resp_data.get("context", {}).get("matched")
        if not resp_data:
            return []
        if self._list_key in resp_data:
            resp_data = resp_data[self._list_key]

        return [
            self._cls(
                self._client,
                headers,
                element,
                parent=self._parent,
            )
            for element in resp_data
            if element is not None
        ]

    def get_limited(self) -> list[T]:
        """Returns the first page of the results, up to pagination limit

        Returns:
            list[T]: List of objects
        """

        limit = consts.PAGINATION_LIMIT
        if self._data:
            limit: int = self._data.get("per_page", limit)

        if len(self._elements) < limit:
            self._fetch_to_index(limit)
        return self._elements[:limit]

    class PagedSlice:
        def __init__(self, _list: PaginatedList[T], _slice: slice) -> None:
            self._list = _list
            self._start = _slice.start or 0
            self._stop = _slice.stop
            self._step = _slice.step or 1

        def __iter__(self) -> Iterator[T]:
            index = self._start
            while not (self._stop is not None and index >= self._stop):
                if len(self._list._elements) > index or self._list._has_next():
                    yield self._list[index]
                    index += self._step
                else:
                    return

        def __getitem__(self, index: int | slice) -> T:
            return self._list[index]



from dataclasses import dataclass
from typing import Literal, TypedDict, TypeVar

from requests.structures import CaseInsensitiveDict

Headers = CaseInsensitiveDict
Params = dict[str, str | int]
RequestMethod = Literal["GET", "POST", "DELETE", "PUT"]


class SearchSortField(TypedDict):
    field: str
    direction: Literal["asc", "desc"]


class SearchFields(TypedDict):
    include: list[str]
    exclude: list[str]


L = TypeVar("L", bound="Link")


@dataclass
class Link:
    """A class representing a hypermedia link.

    Attributes:
        rel: The relationship type of the link.
        href: The URL the link points to.
        title: Optional title of the link.
        media_type: Optional media type of the linked resource.
    """

    rel: str
    href: str
    title: str | None = None
    media_type: str | None = None

    @classmethod
    def from_dict(cls: type[L], data: dict[str, str]) -> L:
        """Create a Link instance from a dictionary.

        Args:
            data: Dictionary containing link data.

        Returns:
            A new Link instance.
        """
        return cls(
            rel=data["rel"],
            href=data["href"],
            title=data.get("title"),
            media_type=data.get("type"),
        )

    @staticmethod
    def get_link(links: list[L], rel: str) -> L | None:
        """Find a link with the specified relationship type.

        Args:
            links: List of Link objects to search through.
            rel: The relationship type to look for.

        Returns:
            The matching Link object or None if not found.
        """
        return next((ln for ln in links if rel == ln.rel), None)



from unittest.mock import Mock

import pytest

from pyeodh.eodh_object import EodhObject
from pyeodh.pagination import PaginatedList


class DummyItem(EodhObject):
    """A simple test object class inheriting from EodhObject."""

    def __init__(self, client, headers, data, parent=None):
        super().__init__(client, headers, data, parent)

    def _set_props(self, obj: dict) -> None:
        self.id = obj.get("id")


@pytest.fixture
def mock_client():
    """Fixture to create a mock client."""
    return Mock()


@pytest.fixture
def paginated_list(mock_client):
    """Fixture to create a PaginatedList instance."""
    return PaginatedList(
        DummyItem,
        mock_client,
        "GET",
        "https://api.example.com/items",
        "items",
        params={"param1": "value1"},
    )


def test_get_limited_first_page(paginated_list, mock_client):
    """Test get_limited() when fetching the first page of results."""
    mock_response = {
        "items": [{"id": i} for i in range(10)],
        "links": [{"rel": "next", "href": "https://api.example.com/items?page=2"}],
        "context": {"matched": 25},
    }
    mock_client._request_json.return_value = ({}, mock_response)

    result = paginated_list.get_limited()

    assert len(result) == 10
    assert all(isinstance(item, DummyItem) for item in result)
    assert [item.id for item in result] == list(range(10))


@pytest.mark.parametrize(
    "a_slice",
    [
        slice(0, 5, None),
        slice(2, 7, None),
        slice(0, 10, None),
        slice(5, 10, None),
        slice(0, 10, 2),
        slice(1, 8, 3),
    ],
)
def test_pagination_slicing_support(
    paginated_list: PaginatedList,
    mock_client: Mock,
    a_slice: slice,
) -> None:
    """Test that slicing is supported with different slice combinations."""

    mock_response = {
        "items": [{"id": i} for i in range(10)],
        "links": [{"rel": "next", "href": "https://api.example.com/items?page=2"}],
        "context": {"matched": 25},
    }
    mock_client._request_json.return_value = ({}, mock_response)

    result = list(paginated_list[a_slice])

    range_obj = range(*a_slice.indices(25))
    slice_length = len(range_obj)

    assert len(result) == slice_length
    assert [item.id for item in result] == list(range_obj)


def test_get_limited_partial_page(paginated_list, mock_client):
    """Test get_limited() when fetching a partial page of results."""
    mock_response = {
        "items": [{"id": i} for i in range(5)],
        "links": [],
        "context": {"matched": 5},
    }
    mock_client._request_json.return_value = ({}, mock_response)

    result = paginated_list.get_limited()

    assert len(result) == 5
    assert all(isinstance(item, DummyItem) for item in result)
    assert [item.id for item in result] == list(range(5))


def test_get_limited_custom_limit(mock_client):
    """Test get_limited() with a custom pagination limit."""
    custom_limit = 15
    paginated_list = PaginatedList(
        DummyItem,
        mock_client,
        "GET",
        "https://api.example.com/items",
        "items",
        params={"param1": "value1"},
        first_data={"per_page": custom_limit},
    )

    mock_response = {
        "items": [{"id": i} for i in range(custom_limit)],
        "links": [{"rel": "next", "href": "https://api.example.com/items?page=2"}],
        "context": {"matched": 30},
    }
    mock_client._request_json.return_value = ({}, mock_response)

    result = paginated_list.get_limited()

    assert len(result) == custom_limit
    assert all(isinstance(item, DummyItem) for item in result)
    assert [item.id for item in result] == list(range(custom_limit))


def test_get_limited_cached_results(paginated_list, mock_client):
    """Test get_limited() when results are already cached."""
    # Populate the _elements list
    paginated_list._elements = [DummyItem(mock_client, {}, {"id": i}) for i in range(10)]

    result = paginated_list.get_limited()

    assert len(result) == 10
    assert all(isinstance(item, DummyItem) for item in result)
    assert [item.id for item in result] == list(range(10))
    mock_client._request_json.assert_not_called()


def test_iter_method(mock_client):
    """Test the __iter__ method of PaginatedList."""
    paginated_list = PaginatedList(
        DummyItem,
        mock_client,
        "GET",
        "https://api.example.com/items",
        "items",
        params={"param1": "value1"},
    )

    # Mock responses for multiple pages
    mock_responses = [
        {
            "items": [{"id": i} for i in range(5)],
            "links": [{"rel": "next", "href": "https://api.example.com/items?page=2"}],
            "context": {"matched": 15},
        },
        {
            "items": [{"id": i} for i in range(5, 10)],
            "links": [{"rel": "next", "href": "https://api.example.com/items?page=3"}],
            "context": {"matched": 15},
        },
        {
            "items": [{"id": i} for i in range(10, 15)],
            "links": [],
            "context": {"matched": 15},
        },
    ]

    mock_client._request_json.side_effect = [({}, response) for response in mock_responses]

    # Use the iterator to fetch all items
    items = list(paginated_list)

    assert len(items) == 15
    assert all(isinstance(item, DummyItem) for item in items)
    assert [item.id for item in items] == list(range(15))

    # Check that _request_json was called for each page
    assert mock_client._request_json.call_count == 3


def test_iter_method_empty_response(mock_client):
    """Test the __iter__ method of PaginatedList with an empty response."""
    paginated_list = PaginatedList(
        DummyItem,
        mock_client,
        "GET",
        "https://api.example.com/items",
        "items",
        params={"param1": "value1"},
    )

    mock_response = {
        "items": [],
        "links": [],
        "context": {"matched": 0},
    }

    mock_client._request_json.return_value = ({}, mock_response)

    items = list(paginated_list)

    assert len(items) == 0
    mock_client._request_json.assert_called_once()


def test_iter_method_single_page(mock_client):
    """Test the __iter__ method of PaginatedList with a single page response."""
    paginated_list = PaginatedList(
        DummyItem,
        mock_client,
        "GET",
        "https://api.example.com/items",
        "items",
        params={"param1": "value1"},
    )

    mock_response = {
        "items": [{"id": i} for i in range(5)],
        "links": [],
        "context": {"matched": 5},
    }

    mock_client._request_json.return_value = ({}, mock_response)

    items = list(paginated_list)

    assert len(items) == 5
    assert all(isinstance(item, DummyItem) for item in items)
    assert [item.id for item in items] == list(range(5))
    mock_client._request_json.assert_called_once()


def test_fetch_next_raises_runtime_error(mock_client):
    paginated_list = PaginatedList(
        DummyItem,
        mock_client,
        "GET",
        "https://api.example.com/items",
        "items",
        params={"param1": "value1"},
    )

    # Set _next_url to None to simulate missing next URL
    paginated_list._next_url = None

    with pytest.raises(RuntimeError, match="Next url not specified!"):
        paginated_list._fetch_next()

    mock_client._request_json.assert_not_called()



from unittest.mock import patch

import pytest
from pystac import Extent

import pyeodh
import pyeodh.pagination
import pyeodh.resource_catalog
from pyeodh import consts
from pyeodh.resource_catalog import CatalogService, Item
from pyeodh.utils import ConformanceError

CEDA_CAT_ID = "public/catalogs/ceda-stac-catalogue"


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["Authorization"],
        "decode_compressed_response": True,
    }


@pytest.fixture
def svc() -> CatalogService:
    return pyeodh.Client(base_url="https://staging.eodatahub.org.uk").get_catalog_service()


@pytest.mark.vcr
def test_get_catalog_service(svc: CatalogService):
    assert svc._pystac_object.STAC_OBJECT_TYPE == "Catalog"
    assert svc._pystac_object.self_href == "https://staging.eodatahub.org.uk/api/catalogue/stac/"


@pytest.mark.vcr
def test_ping(svc: CatalogService):
    msg = svc.ping()
    assert msg == "PONG"


@pytest.mark.vcr
def test_get_conformance(svc: CatalogService):
    conformances = svc.get_conformance()
    assert isinstance(conformances, list)
    assert all(isinstance(elem, str) for elem in conformances)


@pytest.mark.vcr
def test_get_catalog(svc: CatalogService):
    cat = svc.get_catalog("public")
    assert cat.id == "public"


@pytest.mark.vcr
def test_get_catalogs_from_catalog(svc: CatalogService):
    parent_catalog = svc.get_catalog("public")
    children = parent_catalog.get_catalogs()
    assert isinstance(children, list)
    assert all(isinstance(elem, pyeodh.resource_catalog.Catalog) for elem in children)


@pytest.mark.vcr
def test_get_collections_from_catalog(svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collections = cat.get_collections()
    assert isinstance(collections, pyeodh.pagination.PaginatedList)
    assert all(isinstance(elem, pyeodh.resource_catalog.Collection) for elem in collections)


@pytest.mark.vcr
def test_get_collection_from_catalog(svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    assert collection.id == "cmip6"


@pytest.mark.vcr
def test_get_collection_items(svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    items = collection.get_items()
    assert isinstance(items, pyeodh.pagination.PaginatedList)

    for i in range(20):
        assert isinstance(items[i], pyeodh.resource_catalog.Item)


@pytest.mark.vcr
def test_get_collection_items_with_limit(svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    items = collection.get_items().get_limited()
    assert len(items) == consts.PAGINATION_LIMIT


@pytest.mark.vcr
def test_get_collection_items_total_count(svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    items = collection.get_items()
    assert isinstance(items.total_count, int)


@pytest.mark.vcr
def test_get_collection_item(svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    items = collection.get_items()
    item = collection.get_item(items[0].id)
    assert isinstance(item, pyeodh.resource_catalog.Item)
    assert item.id == items[0].id


@pytest.mark.skip(reason="ceda-datapoint is broken")
@pytest.mark.vcr
def test_get_cloud_product(svc: CatalogService):
    import xarray
    from ceda_datapoint.core.cloud import DataPointCloudProduct

    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    items = collection.get_items()
    item = collection.get_item(items[0].id)

    product = item.get_cloud_products()
    assert isinstance(product, DataPointCloudProduct)
    assert item.id in product.id

    ds = product.open_dataset()
    assert isinstance(ds, xarray.Dataset)


@pytest.mark.vcr
def test_get_item_from_href(svc: CatalogService):
    item = pyeodh.resource_catalog.Item.from_href(
        svc._client,
        "https://staging.eodatahub.org.uk/api/catalogue/stac/catalogs/public/catalogs/"
        "ceda-stac-catalogue/collections/sentinel2_ard/items/neodc.sentinel_ard.data."
        "sentinel_2.2025.04.11.S2A_20250411_latn572lonw0021_T30VWJ_ORB123_20250411183"
        "553_utm30n_osgb",
    )
    assert isinstance(item, pyeodh.resource_catalog.Item)
    assert (
        item.id == "neodc.sentinel_ard.data.sentinel_2.2025.04.11."
        "S2A_20250411_latn572lonw0021_T30VWJ_ORB123_20250411183553_utm30n_osgb"
    )


@pytest.mark.vcr
@patch(
    "pyeodh.resource_catalog.CatalogService.get_conformance",
    return_value=["https://api.stacspec.org/v1.0.0/core"],
)
def test_conformance_error_raised(mock_get_conformance, svc: CatalogService):
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    item = collection.get_items()[0]
    # Test Catalog methods
    with pytest.raises(ConformanceError):
        svc.create_catalog("test_id", "test_description")
    with pytest.raises(ConformanceError):
        cat.update(description="new description")
    with pytest.raises(ConformanceError):
        cat.delete()

    # Test Collection methods
    with pytest.raises(ConformanceError):
        cat.create_collection(
            "test_id",
            "test_description",
            Extent.from_dict(
                {
                    "spatial": {
                        "bbox": [
                            [
                                -9.00034454651177,
                                49.48562028352171,
                                3.1494256015866995,
                                61.33444247301668,
                            ]
                        ]
                    },
                    "temporal": {"interval": [["2023-01-01T11:14:51Z", "2023-11-01T11:43:49Z"]]},
                }
            ),
        )
    with pytest.raises(ConformanceError):
        collection.update(description="new description")
    with pytest.raises(ConformanceError):
        collection.delete()

    # Test Item methods
    with pytest.raises(ConformanceError):
        collection.create_item("test_id", None, None, None, {})
    with pytest.raises(ConformanceError):
        item.update(properties={"new": "property"})
    with pytest.raises(ConformanceError):
        item.delete()
    # with pytest.raises(ConformanceError):
    #    _ = item.get_cloud_products()


@patch("pyeodh.client.Client._request_json")
def test_from_href_mocked(mock_request_json, svc: CatalogService) -> None:
    """Test creating an Item from a URL using mocked response.

    Args:
        mock_request_json: Mock for the client's _request_json method.
        svc (CatalogService): CatalogService fixture.
    """
    href = "https://example.com/api/stac/item1"

    # Mock response data
    mock_response_data = {
        "type": "Feature",
        "stac_version": "1.0.0",
        "id": "item1",
        "collection": "collection1",
        "geometry": {
            "type": "Polygon",
            "coordinates": [[[0, 90], [0, -90], [-180, -90], [-180, 90], [0, 90]]],
        },
        "bbox": [0, 90, 0, -90],
        "properties": {
            "datetime": "2023-11-20T11:23:51Z",
            "updated": "2024-11-05T07:43:53.975696Z",
            "start_datetime": "2023-11-20T11:23:51Z",
            "end_datetime": "2023-11-20T11:23:51Z",
        },
        "links": [
            {
                "rel": "self",
                "type": "application/geo+json",
                "href": "https://example.com/api/stac/item1",
            }
        ],
        "assets": {},
    }

    # Mock the _request_json method to return the mock response
    mock_request_json.return_value = (None, mock_response_data)

    item = Item.from_href(svc._client, href)
    assert isinstance(item, Item)
    assert item.id == mock_response_data["id"]


# Tests for self_href property and path discovery methods


@pytest.mark.vcr
def test_catalog_service_self_href(svc: CatalogService) -> None:
    """Test that CatalogService exposes self_href property.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    assert svc.self_href is not None
    assert isinstance(svc.self_href, str)
    assert "stac" in svc.self_href


@pytest.mark.vcr
def test_catalog_self_href(svc: CatalogService) -> None:
    """Test that Catalog exposes self_href property.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    cat = svc.get_catalog("public")
    assert cat.self_href is not None
    assert isinstance(cat.self_href, str)
    assert "catalogs/public" in cat.self_href


@pytest.mark.vcr
def test_collection_self_href(svc: CatalogService) -> None:
    """Test that Collection exposes self_href property.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    assert collection.self_href is not None
    assert isinstance(collection.self_href, str)
    assert "collections/cmip6" in collection.self_href


@pytest.mark.vcr
def test_item_self_href(svc: CatalogService) -> None:
    """Test that Item exposes self_href property.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    items = collection.get_items()
    item = items[0]
    assert item.self_href is not None
    assert isinstance(item.self_href, str)
    assert "items/" in item.self_href


@pytest.mark.vcr
def test_catalog_get_path(svc: CatalogService) -> None:
    """Test Catalog.get_path() returns the catalog path.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    cat = svc.get_catalog("public")
    path = cat.get_path()
    assert path == "public"


@pytest.mark.vcr
def test_nested_catalog_get_path(svc: CatalogService) -> None:
    """Test Catalog.get_path() returns nested catalog path.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    cat = svc.get_catalog(CEDA_CAT_ID)
    path = cat.get_path()
    assert path == CEDA_CAT_ID


@pytest.mark.vcr
def test_collection_get_catalog_path(svc: CatalogService) -> None:
    """Test Collection.get_catalog_path() returns the parent catalog path.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    cat = svc.get_catalog(CEDA_CAT_ID)
    collection = cat.get_collection("cmip6")
    catalog_path = collection.get_catalog_path()
    assert catalog_path == CEDA_CAT_ID


@pytest.mark.vcr
def test_get_catalog_paths(svc: CatalogService) -> None:
    """Test CatalogService.get_catalog_paths() returns list of catalog paths.

    Args:
        svc (CatalogService): CatalogService fixture.
    """
    paths = svc.get_catalog_paths(recursive=False)
    assert isinstance(paths, list)
    assert all(isinstance(p, str) for p in paths)
    # Should have at least the 'public' catalog
    assert any("public" in p for p in paths)



from pyeodh.types import Link


def test_link_from_dict() -> None:
    data = {
        "rel": "self",
        "href": "https://example.com",
        "title": "Example",
        "type": "application/json",
    }

    link = Link.from_dict(data)

    assert link.rel == data["rel"]
    assert link.href == data["href"]
    assert link.title == data["title"]
    assert link.media_type == data["type"]


def test_get_link() -> None:
    links = [
        Link(rel="self", href="https://example.com"),
        Link(rel="next", href="https://example.com/next"),
        Link(rel="prev", href="https://example.com/prev"),
    ]

    assert Link.get_link(links, "self") == links[0]
    assert Link.get_link(links, "next") == links[1]
    assert Link.get_link(links, "prev") == links[2]
    assert Link.get_link(links, "nonexistent") is None



import pytest
from owslib.wmts import WebMapTileService

import pyeodh


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["Authorization"],
        "decode_compressed_response": True,
    }


@pytest.mark.vcr
def test_get_wmts() -> None:
    wmts = (
        pyeodh.Client(base_url="https://staging.eodatahub.org.uk")
        .get_catalog_service()
        .get_catalog("public/catalogs/ceda-stac-catalogue")
        .get_wmts()
    )
    assert isinstance(wmts, WebMapTileService)

    assert hasattr(wmts, "contents")
    assert hasattr(wmts, "operations")
    assert hasattr(wmts, "tilematrixsets")

    # Check if the name attribute patch is applied
    for op in wmts.operations:
        assert hasattr(op, "name")



from __future__ import annotations

import logging
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal, TypeVar

import pystac
import pystac.catalog
from ceda_datapoint.core.cloud import DataPointCloudProduct, DataPointCluster
from ceda_datapoint.core.item import identify_cloud_type
from owslib.wmts import WebMapTileService
from pystac import Extent, RelType, STACObject, STACTypeError, Summaries
from pystac.asset import Asset
from pystac.provider import Provider

from pyeodh import consts
from pyeodh.eodh_object import EodhObject, is_optional
from pyeodh.pagination import PaginatedList
from pyeodh.types import Headers, SearchFields, SearchSortField
from pyeodh.utils import ConformanceError, join_url, remove_null_items

if TYPE_CHECKING:
    # avoids conflicts since there are also kwargs and attrs called `datetime`
    from datetime import datetime as Datetime

    from pyeodh.client import Client

logger = logging.getLogger(__name__)


C = TypeVar("C", bound="STACObject")


class Conformance(Enum):
    TRANSACTION_EXTENSION = "https://api.stacspec.org/v1.0.0/ogcapi-features/extensions/transaction"


class Item(EodhObject):
    """Represents a STAC Item in the resource catalog."""

    _pystac_object: pystac.Item

    def __init__(self, client: Client, headers: Headers, data: dict, **kwargs: Any) -> None:
        super().__init__(client, headers, data, pystac.Item, **kwargs)

    def _set_props(self, obj: pystac.Item) -> None:
        self.id = obj.id
        self.geometry = obj.geometry
        self.bbox = obj.bbox
        self.datetime = obj.datetime
        self.properties = obj.properties
        self.collection = obj.collection_id
        self.assets = obj.assets

    @property
    def self_href(self) -> str:
        """URL pointing to this item's endpoint.

        Returns:
            str: The self href URL.
        """
        return self._pystac_object.self_href

    def delete(self) -> None:
        """Delete this item.


        Calls: DELETE /catalogs/{catalog_id}/collections/{collection_id}/items/{item_id}
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        self._client._request_raw("DELETE", self._pystac_object.self_href)

    def update(
        self,
        geometry: dict[str, Any] | None = None,
        bbox: list[float] | None = None,
        datetime: Datetime | None = None,
        properties: dict[str, Any] | None = None,
        collection: str | pystac.Collection | None = None,
        assets: dict[str, Any] | None = None,
    ) -> None:
        """Update this item's attributes with new values. Only provide the values you
        want to update.

        Calls: PUT /catalogs/{catalog_id}/collections/{collection_id}/items/{item_id}


        Args:
            geometry (dict[str, Any] | None, optional): Geometry. Defaults to None.
            bbox (list[float] | None, optional): Bounding box. Defaults to None.
            datetime (Datetime | None, optional): Datetime. Defaults to None.
            properties (dict[str, Any] | None, optional): Properties. Defaults to None.
            collection (str | pystac.Collection | None, optional): Collection. Defaults
                to None.
            assets (dict[str, Any] | None, optional): Assets. Defaults to None.
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )

        put_data = remove_null_items(
            {
                "id": self.id,
                "geometry": geometry or self.geometry,  # ! getting 500 with this
                "bbox": bbox or self.bbox,
                "datetime": datetime or self.datetime,
                "properties": properties or self.properties,
                "collection": collection or self.collection,
                "assets": assets or self.assets,
            }
        )

        _, resp_data = self._client._request_json("PUT", self._pystac_object.self_href, data=put_data)

        if resp_data:
            self._set_props(self._pystac_object.from_dict(resp_data))

    def get_cloud_products(
        self,
    ) -> DataPointCloudProduct | DataPointCluster | list[DataPointCloudProduct]:
        """Retrieve the cloud product(s) attributed to this item.

        Feature added from the CEDA DataPoint library to access cloud
        products as either an individual product object or a set of
        products represented by a cluster. See the documentation
        for using cloud products and clusters at:
        https://cedadev.github.io/datapoint/usage.html

        Typical usage:
            product = item.get_cloud_products()
            ds = product.open_dataset()
            # Continue with the `ds` xarray.Dataset
        """

        products = []
        # Iterate over assets in this item
        for id, asset in self.assets.items():
            cf = identify_cloud_type(id, asset)
            if cf is None:
                continue

            products.append(
                DataPointCloudProduct(
                    asset.to_dict(),  # Convert Asset to dict
                    id=f"{self.id}-{id}",
                    cf=cf,
                    meta={"bbox": self.bbox},
                    properties=self.properties,
                )
            )

        if len(products) == 0:
            return []
        if len(products) == 1:
            return products[0]

        # Could also just return the list.
        # See https://cedadev.github.io/datapoint/cloud_formats.html
        # for reasons to use the cluster.

        return DataPointCluster(products, parent_id=f"{self.id}-cluster", meta={"bbox": self.bbox})

    def commercial_data_order(
        self,
        product_bundle: str,
        workspace: str | None = None,
        aoi: list[list[list[float]]] | None = None,
        extra_data: dict | None = None,
    ) -> None:
        """Order commercial data. Available only for specific catalogs.


        Args:
            workspace (str, optional): The workspace to order the data from. Defaults to
                the username pyeodh client was initialized with.
            product_bundle (str): The product bundle to order from the commercial data
                provider.
            aoi (list[float] | None, optional): Coordinates to limit the AOI of the item
                for purchase where possible. Given in the same nested format as STAC.
                Defaults to None.
            extra_data (dict | None, optional): A placeholder for future data options to
                include in the item. Defaults to None.
        """

        if workspace is None:
            workspace = self._client.username

        api_url = f"/api/catalogue/manage/catalogs/user-datasets/{workspace}/commercial-data"
        payload: dict[str, Any] = {
            "url": self._pystac_object.self_href,
            "product_bundle": product_bundle,
        }
        if aoi:
            payload["coordinates"] = aoi
        if extra_data:
            payload["extra_data"] = extra_data

        _, data = self._client._request_json("POST", api_url, data=payload)

        logger.info(f"Commercial data ordered: {data.get('message')}")


class Collection(EodhObject):
    """Represents a STAC Collection in the resource catalog."""

    _pystac_object: pystac.Collection

    def __init__(self, client: Client, headers: Headers, data: dict, **kwargs: Any) -> None:
        super().__init__(client, headers, data, pystac.Collection, **kwargs)

    def _set_props(self, obj: pystac.Collection) -> None:
        self.id = obj.id
        self.description = obj.description
        self.extent = obj.extent
        self.title = obj.title
        self.license = obj.license
        self.keywords = obj.keywords
        self.providers = obj.providers
        self.summaries = obj.summaries
        self.assets = obj.assets

    @property
    def self_href(self) -> str:
        """URL pointing to this collection's endpoint.

        Returns:
            str: The self href URL.
        """
        return self._pystac_object.self_href

    @cached_property
    def items_href(self) -> str:
        """URL pointing to items endpoint."""
        link = self._pystac_object.get_single_link(RelType.ITEMS)
        if not link:
            raise RuntimeError("Object does not have items link!")
        return link.href

    def get_catalog_path(self) -> str:
        """Get the parent catalog path for this collection.

        Extracts the catalog path from the collection's self_href URL.
        For example, if self_href is:
        `.../catalogs/public/catalogs/ceda-stac-catalogue/collections/cmip6`
        This returns: `public/catalogs/ceda-stac-catalogue`

        Returns:
            str: The parent catalog path.

        Raises:
            ValueError: If the catalog path cannot be extracted from the URL.
        """
        href = self.self_href
        # Find the catalogs segment and extract path up to /collections
        if "/catalogs/" not in href or "/collections/" not in href:
            raise ValueError(f"Cannot extract catalog path from URL: {href}")

        # Find the first /catalogs/ and extract until /collections/
        catalogs_idx = href.find("/catalogs/")
        collections_idx = href.find("/collections/")

        if catalogs_idx == -1 or collections_idx == -1:
            raise ValueError(f"Cannot extract catalog path from URL: {href}")

        # Extract the path between /catalogs/ and /collections/
        catalog_path = href[catalogs_idx + len("/catalogs/") : collections_idx]
        return catalog_path

    def get_items(self) -> PaginatedList[Item]:
        """Fetches all items within a collection.

        Calls: GET /catalogs/{catalog_id}/collections/{collection_id}/items

        Returns:
            PaginatedList[Item]: Iterable list of items. Automatically handles
                paginated results.
        """
        return PaginatedList(
            Item,
            self._client,
            "GET",
            self.items_href,
            "features",
            params={"limit": consts.PAGINATION_LIMIT},
            parent=self,
        )

    def get_item(self, item_id: str) -> Item:
        """Fetches a collection item.

        Calls: GET /catalogs/{catalog_id}/collections/{collection_id}/items/{item_id}

        Args:
            item_id (str): ID of a collection item.

        Returns:
            Item: Initialized item object.
        """
        url = join_url(self.items_href, item_id)
        headers, response = self._client._request_json("GET", url)

        return Item(self._client, headers, response, parent=self)

    def update(
        self,
        description: str | None = None,
        extent: Extent | None = None,
        title: str | None = None,
        license: str | None = None,
        keywords: list[str] | None = None,
        providers: list[Provider] | None = None,
        summaries: Summaries | None = None,
        assets: dict[str, Asset] | None = None,
    ) -> None:
        """Update collection attributes with new values. Provide only the values you
        want to update.

        Calls: PUT /catalogs/{catalog_id}/collections/{collection_id}


        Args:
            description (str | None, optional): Description. Defaults to None.
            extent (Extent | None, optional): Extent. Defaults to None.
            title (str | None, optional): Title. Defaults to None.
            license (str | None, optional): License. Defaults to None.
            keywords (list[str] | None, optional): Keywords. Defaults to None.
            providers (list[Provider] | None, optional): Providers. Defaults to None.
            summaries (Summaries | None, optional): Summaries. Defaults to None.
            assets (dict[str, Asset] | None, optional): Assets. Defaults to None.
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        assert is_optional(description, str), description
        assert is_optional(extent, Extent), extent
        assert is_optional(title, str), title
        assert is_optional(license, str), license
        assert is_optional(keywords, list), keywords
        assert is_optional(providers, list), providers
        assert is_optional(summaries, Summaries), summaries
        assert is_optional(assets, dict), assets

        put_data = remove_null_items(
            {
                "id": self.id,
                "description": description or self.description,
                "extent": extent or self.extent,
                "title": title or self.title,
                "license": license or self.license,
                "keywords": keywords or self.keywords,
                "providers": providers or self.providers,
                "summaries": summaries or self.summaries,
                "assets": assets or self.assets,
            }
        )

        _, resp_data = self._client._request_json("PUT", self._pystac_object.self_href, data=put_data)

        if resp_data:
            self._set_props(self._pystac_object.from_dict(resp_data))

    def delete(self) -> None:
        """Delete this collection.

        Calls: DELETE /catalogs/{catalog_id}/collections/{collection_id}
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        self._client._request_raw("DELETE", self._pystac_object.self_href)

    def create_item(
        self,
        id: str,
        geometry: dict[str, Any] | None,
        bbox: list[float] | None,
        datetime: Datetime | None,
        properties: dict[str, Any] | None,
        collection: str | pystac.Collection | None = None,
        assets: dict[str, Any] | None = None,
    ) -> Item:
        """Create an item as part of this collection.

        Calls: POST /catalogs/{catalog_id}/collections/{collection_id}/items/{item_id}

        Args:
            id (str): A unique ID.
            geometry (dict[str, Any] | None): Geometry
            bbox (list[float] | None): Bounding box
            datetime (Datetime | None): Datetime
            properties (dict[str, Any] | None): Properties
            collection (str | pystac.Collection | None, optional): ID of a parent
                collection. Defaults to None.
            assets (dict[str, Any] | None, optional): Assets. Defaults to None.

        Returns:
            Item: _description_
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        post_data = remove_null_items(
            {
                "id": id,
                "geometry": geometry,
                "bbox": bbox,
                "datetime": datetime,
                "properties": properties,
                "collection": collection,
                "assets": assets,
            }
        )

        headers, response = self._client._request_json("POST", self.items_href, data=post_data)
        return Item(self._client, headers, response, parent=self)


class Catalog(EodhObject):
    """Represents a STAC Catalog in the resource catalog."""

    _pystac_object: pystac.Catalog

    def __init__(self, client: Client, headers: Headers, data: dict, **kwargs: Any) -> None:
        super().__init__(client, headers, data, pystac.Catalog, **kwargs)

    def _set_props(self, obj: pystac.Catalog) -> None:
        self.id = obj.id
        self.description = obj.description
        self.title = obj.title
        self.conforms_to = obj.extra_fields.get("conformsTo", []).copy()

    @property
    def self_href(self) -> str:
        """URL pointing to this catalog's endpoint.

        Returns:
            str: The self href URL.
        """
        return self._pystac_object.self_href

    def get_path(self) -> str:
        """Get the path for this catalog.

        Extracts the catalog path from the catalog's self_href URL.
        For example, if self_href is:
        `.../catalogs/public/catalogs/ceda-stac-catalogue`
        This returns: `public/catalogs/ceda-stac-catalogue`

        Returns:
            str: The catalog path.

        Raises:
            ValueError: If the catalog path cannot be extracted from the URL.
        """
        href = self.self_href
        # Find the catalogs segment and extract the path
        if "/catalogs/" not in href:
            raise ValueError(f"Cannot extract catalog path from URL: {href}")

        # Find the first /catalogs/ and extract everything after it
        catalogs_idx = href.find("/catalogs/")
        catalog_path = href[catalogs_idx + len("/catalogs/") :]

        # Remove trailing slash if present
        catalog_path = catalog_path.rstrip("/")

        return catalog_path

    @cached_property
    def collections_href(self) -> str:
        """URL pointing to collections endpoint."""
        return join_url(self._pystac_object.self_href, "collections")

    @cached_property
    def wmts_href(self) -> str:
        """URL pointing to WMTS endpoint."""
        return join_url(self._pystac_object.self_href, "wmts")

    def get_catalogs(self) -> list[Catalog]:
        """Fetches children catalogs of this parent catalog.

        Calls: GET /catalogs/{catalog_path}/catalogs

        Returns:
            list[Catalog]: List of children catalogs.
        """
        url = join_url(self._pystac_object.self_href, "catalogs")
        headers, data = self._client._request_json("GET", url)
        catalogs = []

        for cat in data.get("catalogs"):
            try:
                catalogs.append(Catalog(self._client, headers, cat, parent=self))
            except STACTypeError as e:
                logger.warning(f"{e} => Skipping")
        return catalogs

    def get_collections(self) -> PaginatedList[Collection]:
        """Fetches all resource catalog collections.

        Calls: GET /catalogs/{catalog_id}/collections

        Returns:
            list[Collection]: List of available collections
        """

        return PaginatedList(
            Collection,
            self._client,
            "GET",
            self.collections_href,
            "collections",
            parent=self,
        )

    def get_collection(self, collection_id: str) -> Collection:
        """Fetches a resource catalog collection.

        Calls: GET /catalogs/{catalog_id}/collections/{collection_id}

        Args:
            collection_id (str): ID of a collection

        Returns:
            Collection: Collection for given ID
        """
        url = join_url(self.collections_href, collection_id)
        headers, response = self._client._request_json("GET", url)

        return Collection(self._client, headers, response, parent=self)

    def create_collection(
        self,
        id: str,
        description: str,
        extent: Extent,
        title: str | None = None,
        license: str | None = None,
        keywords: list[str] | None = None,
        providers: list[Provider] | None = None,
        summaries: Summaries | None = None,
        assets: dict[str, Asset] | None = None,
    ) -> Collection:
        """Create a new collection inside a catalog.

        Calls: POST /catalogs/{catalog_id}/collections

        Args:
            id (str)
            description (str)
            extent (Extent)
            title (str | None, optional): Defaults to None.
            license (str | None, optional): Defaults to None.
            keywords (list[str] | None, optional): Defaults to None.
            providers (list[Provider] | None, optional): Defaults to None.
            summaries (Summaries | None, optional): Defaults to None.
            assets (dict[str, Asset] | None, optional): Defaults to None.

        Returns:
            Collection: An initialized collection object.
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        assert isinstance(id, str), id
        assert isinstance(description, str), description
        assert isinstance(extent, Extent), extent
        assert is_optional(title, str), title
        assert is_optional(license, str), license
        assert is_optional(keywords, list), keywords
        assert is_optional(providers, list), providers
        assert is_optional(summaries, Summaries), summaries
        assert is_optional(assets, dict), assets

        post_data = remove_null_items(
            {
                "id": id,
                "description": description,
                "extent": extent,
                "title": title,
                "license": license,
                "keywords": keywords,
                "providers": providers,
                "summaries": summaries,
                "assets": assets,
            }
        )
        headers, response = self._client._request_json("POST", self.collections_href, data=post_data)
        return Collection(self._client, headers, response, parent=self)

    def update(
        self,
        description: str | None = None,
        title: str | None = None,
    ) -> None:
        """Updates catalog.

        Calls: PUT /catalogs/{catalog_id}

        Args:
            description (str | None, optional): New description.  Defaults to None.
            title (str | None, optional): New title. Defaults to None.
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        assert is_optional(description, str), description
        assert is_optional(title, str), title

        put_data = remove_null_items(
            {
                "id": self.id,
                "description": description or self.description,
                "title": title or self.title,
            }
        )
        _, resp_data = self._client._request_json("PUT", self._pystac_object.self_href, data=put_data)

        if resp_data:
            self._set_props(self._pystac_object.from_dict(resp_data))

    def delete(self) -> None:
        """Delete this catalog.

        Calls: DELETE /catalogs/{catalog_id}
        """
        if not self.get_root().check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        self._client._request_raw("DELETE", self._pystac_object.self_href)

    def search(
        self,
        limit: int = consts.PAGINATION_LIMIT,
        collections: list[str] | None = None,
        catalog_paths: list[str] | None = None,
        ids: list[str] | None = None,
        bbox: list[Any] | None = None,
        intersects: dict | None = None,
        datetime: str | None = None,
        start_datetime: str | None = None,
        end_datetime: str | None = None,
        fields: SearchFields | None = None,
        query: dict[str, Any] | list[str] | None = None,
        sort_by: list[SearchSortField] | None = None,
        filter: dict | None = None,
        filter_crs: str | None = None,
        filter_lang: Literal["cql-json", "cql2-json", "cql2-text"] | None = None,
    ) -> PaginatedList[Item]:
        """Search for items in the catalog.

        Args:
            limit (int): Number of results per page. Defaults to PAGINATION_LIMIT.
            collections (list[str] | None): Filter by collection IDs.
            catalog_paths (list[str] | None): Filter by catalog paths.
            ids (list[str] | None): Filter by item IDs.
            bbox (list[Any] | None): Bounding box filter.
            intersects (dict | None): GeoJSON geometry to intersect with.
            datetime (str | None): Datetime filter string.
            start_datetime (str | None): Filter items with datetime >= this value.
                Format: ISO 8601 (e.g., "2024-01-01" or "2024-01-01T00:00:00Z").
            end_datetime (str | None): Filter items with datetime <= this value.
                Format: ISO 8601 (e.g., "2024-12-31" or "2024-12-31T23:59:59Z").
            fields (SearchFields | None): Fields to include/exclude.
            query (dict[str, Any] | list[str] | None): Query parameters.
            sort_by (list[SearchSortField] | None): Sort order.
            filter (dict | None): CQL2 filter expression.
            filter_crs (str | None): CRS for filter geometries.
            filter_lang (Literal["cql-json", "cql2-json", "cql2-text"] | None):
                Filter language.

        Returns:
            PaginatedList[Item]: Paginated list of matching items.
        """
        assert isinstance(limit, int), limit
        assert is_optional(collections, list), collections
        assert is_optional(catalog_paths, list), catalog_paths
        assert is_optional(ids, list), ids
        assert is_optional(bbox, list), bbox
        # assert is_optional(intersects, dict), intersects
        assert is_optional(datetime, str), datetime
        assert is_optional(start_datetime, str), start_datetime
        assert is_optional(end_datetime, str), end_datetime
        assert is_optional(fields, dict), fields
        assert is_optional(query, (dict, list)), query
        assert is_optional(sort_by, list), sort_by
        assert is_optional(filter, dict), filter
        assert is_optional(filter_crs, str), filter_crs
        assert filter_lang in ["cql-json", "cql2-json", "cql2-text", None]

        if self.id == "planet" and catalog_paths is not None:
            logger.warning("Catalog paths are not supported for planet catalog, ignoring.")
            catalog_paths = None

        # Build datetime filter if start_datetime or end_datetime provided
        if start_datetime is not None or end_datetime is not None:
            datetime_filter_args = []
            if start_datetime is not None:
                datetime_filter_args.append(
                    {
                        "op": ">=",
                        "args": [
                            {"property": "properties.datetime"},
                            start_datetime,
                        ],
                    }
                )
            if end_datetime is not None:
                datetime_filter_args.append(
                    {
                        "op": "<=",
                        "args": [{"property": "properties.datetime"}, end_datetime],
                    }
                )

            datetime_filter = (
                {"op": "and", "args": datetime_filter_args}
                if len(datetime_filter_args) > 1
                else datetime_filter_args[0]
            )

            # Merge with existing filter if provided
            if filter is not None:
                filter = {"op": "and", "args": [filter, datetime_filter]}
            else:
                filter = datetime_filter

        data = remove_null_items(
            {
                "limit": limit,
                "catalog_paths": catalog_paths,
                "collections": collections,
                "ids": ids,
                "bbox": bbox,
                "intersects": self._format_intersects(intersects),
                "datetime": datetime,
                "fields": fields,
                "query": self._make_query_dict(query),
                "sortby": sort_by,
                "filter": filter,
                "filter_crs": filter_crs,
                "filter_lang": filter_lang,
            }
        )
        url = join_url(self._pystac_object.self_href, "search")
        return PaginatedList(Item, self._client, "POST", url, "features", first_data=data, parent=self)

    def get_wmts(self) -> WebMapTileService:
        """Initializes the OWSLib WebMapTileService

        Returns:
            WebMapTileService: Initialized WMTS
        """
        wmts = WebMapTileService(self.wmts_href)

        # Patch wmts object attribute error
        # see https://github.com/geopython/OWSLib/issues/572
        for i, op in enumerate(wmts.operations):
            if not hasattr(op, "name"):
                wmts.operations[i].name = ""

        return wmts


class CatalogService(Catalog):
    """Represents the root STAC Catalog Service."""

    def get_collections(self) -> PaginatedList[Collection]:
        """Fetches all resource catalog collections.

        Calls: GET /collections

        Returns:
            list[Collection]: List of available collections
        """
        return super().get_collections()

    def get_catalog_paths(self, recursive: bool = True) -> list[str]:
        """Get all catalog paths available in the catalog service.

        Discovers and returns the paths of all catalogs. When recursive is True,
        it traverses the entire catalog hierarchy to find all nested catalogs.

        Args:
            recursive (bool): If True, recursively discover nested catalogs.
                Defaults to True.

        Returns:
            list[str]: List of catalog paths (e.g., ['public', 'public/catalogs/ceda']).
        """
        paths: list[str] = []

        def _collect_paths(catalog: Catalog) -> None:
            """Recursively collect catalog paths."""
            try:
                path = catalog.get_path()
                paths.append(path)
            except ValueError:
                # Skip catalogs where path cannot be extracted
                pass

            if recursive:
                try:
                    children = catalog.get_catalogs()
                    for child in children:
                        _collect_paths(child)
                except Exception as e:
                    logger.warning(f"Failed to get child catalogs: {e}")

        # Start with top-level catalogs
        top_level_catalogs = self.get_catalogs()
        for catalog in top_level_catalogs:
            _collect_paths(catalog)

        return paths

    def get_catalog(self, catalog_id: str) -> Catalog:
        """Fetches a catalog.

        Calls: GET /catalogs/{catalog_id}

        Args:
            catalog_id (str): Catalog ID

        Returns:
            Catalog: An initialized resource catalog object.
        """

        url = join_url(self._pystac_object.self_href, "catalogs", catalog_id)
        headers, data = self._client._request_json("GET", url)
        return Catalog(self._client, headers, data, parent=self)

    def get_catalogs(self) -> list[Catalog]:
        """Fetches all catalogs.

        Calls: GET /catalogs

        Returns:
            list[Catalog]: List of all catalogs available.
        """
        return super().get_catalogs()

    def create_catalog(
        self,
        id: str,
        description: str,
        title: str | None = None,
    ) -> Catalog:
        """Creates a new catalog

        Calls: POST /catalogs

        Args:
            id (str): New catalog ID
            description (str): Catalog description
            title (str | None, optional): Catalog title. Defaults to None.

        Returns:
            Catalog: An initialized catalog object.
        """
        if not self.check_conforms_to(Conformance.TRANSACTION_EXTENSION.value):
            raise ConformanceError(
                f"{Conformance.TRANSACTION_EXTENSION.value}",
            )
        assert isinstance(id, str), id
        assert isinstance(description, str), description
        assert is_optional(title, str), title

        post_data = remove_null_items(
            {
                "id": id,
                "description": description,
                "title": title,
            }
        )
        headers, response = self._client._request_json("POST", self.collections_href, data=post_data)
        return Catalog(self._client, headers, response, parent=self)

    def collection_search(
        self,
        limit: int = consts.PAGINATION_LIMIT,
        bbox: list[Any] | None = None,
        datetime: str | None = None,
        query: str | None = None,
    ) -> PaginatedList[Collection]:
        """Searches the catalog for collections.

        Args:
            limit (int, optional): Number of results per page. Defaults to
                consts.PAGINATION_LIMIT.
            bbox (list[Any] | None, optional): Bounding box. Defaults to None.
            datetime (str | None, optional): Datetime. Defaults to None.
            query (str | None, optional): Query string. Defaults to None.

        Returns:
            PaginatedList[Collection]: Iterable list of collections.
        """

        assert isinstance(limit, int), limit
        assert is_optional(bbox, list), bbox
        assert is_optional(datetime, str), datetime
        assert is_optional(query, str), query

        data = remove_null_items(
            {
                "limit": limit,
                "bbox": bbox,
                "datetime": datetime,
                "q": query,
            }
        )
        url = join_url(self._pystac_object.self_href, "collection-search")
        return PaginatedList(
            Collection,
            self._client,
            "POST",
            url,
            "collections",
            first_data=data,
            parent=self,
        )

    def discovery_search(
        self,
        query: str,
        limit: int = consts.PAGINATION_LIMIT,
    ) -> PaginatedList[Catalog]:
        """Searches the catalog for catalogs and collections.

        Args:
            limit (int, optional): Number of results per page. Defaults to
                consts.PAGINATION_LIMIT.
            query (str | None, optional): Query string. Defaults to None.

        Returns:
            PaginatedList[Collection]: Iterable list of collections.
        """

        assert isinstance(query, str), query
        assert isinstance(limit, int), limit

        data = remove_null_items(
            {
                "limit": limit,
                "q": query,
            }
        )
        url = join_url(self._pystac_object.self_href, "discovery-search")

        return PaginatedList(
            Catalog,
            self._client,
            "POST",
            url,
            "catalogs_and_collections",
            first_data=data,
            parent=self,
        )

    def get_conformance(self) -> list[str]:
        """Fetches list of standards the API conforms to.

        Calls: GET /conformance

        Returns:
            list[str]: Standards.
        """
        url = join_url(self._pystac_object.self_href, "conformance")
        _, response = self._client._request_json("GET", url)
        return response.get("conformsTo", [])

    def ping(self) -> str | None:
        """Pings API.

        Calls: GET /_mgmt/ping

        Returns:
            str | None: Pong.
        """
        _, response = self._client._request_json("GET", join_url(self._pystac_object.self_href, "_mgmt/ping"))
        return response.get("message")

    def check_conforms_to(self, conformance_uri: str | Conformance) -> bool:
        """Test if API conforms to a specification.

        Args:
            conformance_uri (str | Conformance): URI of the specification

        Returns:
            bool
        """
        return conformance_uri in self.get_conformance()



import time

import pytest
import requests

import pyeodh
from pyeodh.ades import Ades, AdesJobStatus, Job, Process, ResultsNotReadyError
from pyeodh.pagination import PaginatedList
from pyeodh.resource_catalog import Item


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["Authorization"],
        "decode_compressed_response": True,
    }


@pytest.fixture
def svc(api_token: str, username: str) -> Ades:
    return pyeodh.Client(username=username, token=api_token, base_url="https://staging.eodatahub.org.uk").get_ades()


@pytest.mark.vcr
def test_get_ades_service(svc: Ades):
    assert (
        svc.self_href == "https://staging.eodatahub.org.uk/api/catalogue/stac/catalogs/user/catalogs/"
        f"{svc._client.username}"
    )


@pytest.mark.vcr
def test_get_processes(svc: Ades):
    processes = svc.get_processes()
    assert isinstance(processes, list)
    assert all(isinstance(p, Process) for p in processes)


@pytest.mark.vcr
def test_get_non_existent_process(svc: Ades):
    with pytest.raises(requests.exceptions.HTTPError):
        svc.get_process("non-existent-process-id")


@pytest.mark.vcr
def test_process_execution(svc: Ades):

    try:
        svc.get_process("convert-url").delete()
    except requests.exceptions.HTTPError:
        pass

    # Deploy a process
    process = svc.deploy_process(
        cwl_url=(
            "https://raw.githubusercontent.com/EOEPCA/deployment-guide/"
            "caaf6eb539c41ff04c7c2b47a1e01b5b7fa6e87f/deploy/samples/requests/"
            "processing/convert-url-app.cwl"
        )
    )
    assert isinstance(process, Process)
    assert process.id == "convert-url"

    # Get a signle process
    single_process = svc.get_process("convert-url")
    assert isinstance(single_process, Process)
    assert single_process.id == "convert-url"

    # Update a process
    cwl_yaml = """cwlVersion: v1.0
$namespaces:
  s: https://schema.org/
s:softwareVersion: 0.1.2
schemas:
  - http://schema.org/version/9.0/schemaorg-current-http.rdf
$graph:
  # Workflow entrypoint
  - class: Workflow
    id: convert-url
    label: convert url app
    doc: Convert URL YAML
    requirements:
      ResourceRequirement:
        coresMax: 1
        ramMax: 1024
    inputs:
      fn:
        label: the operation to perform
        doc: the operation to perform
        type: string
      url:
        label: the image to convert
        doc: the image to convert
        type: string
      size:
        label: the percentage for a resize operation
        doc: the percentage for a resize operation
        type: string
    outputs:
      - id: converted_image
        type: Directory
        outputSource:
          - convert/results
    steps:
      convert:
        run: "#convert"
        in:
          fn: fn
          url: url
          size: size
        out:
          - results
  # convert.sh - takes input args `--url`
  - class: CommandLineTool
    id: convert
    requirements:
      ResourceRequirement:
        coresMax: 1
        ramMax: 512
    hints:
      DockerRequirement:
        dockerPull: eoepca/convert:latest
    baseCommand: convert.sh
    inputs:
      fn:
        type: string
        inputBinding:
          position: 1
      url:
        type: string
        inputBinding:
          position: 2
          prefix: --url
      size:
        type: string
        inputBinding:
          position: 3
    outputs:
      results:
        type: Directory
        outputBinding:
          glob: .
"""

    process.update(cwl_yaml=cwl_yaml)
    assert process.title == "convert url app"
    assert process.description == "Convert URL YAML"

    # Execute a process
    job = process.execute(
        {
            "fn": "resize",
            "url": "https://eoepca.org/media_portal/images/logo6_med.original.png",
            "size": "50%",
        }
    )
    assert isinstance(job, Job)
    assert job.status in [e.value for e in AdesJobStatus]
    assert job.process_id == "convert-url"

    # Refresh a job until finished
    job.refresh()
    while job.status == AdesJobStatus.RUNNING.value:
        time.sleep(2)
        job.refresh()

    # Get result items
    for _ in range(12):
        try:
            items = job.get_result_items()
            if isinstance(items, list) and all(isinstance(item, Item) for item in items):
                break
        except ResultsNotReadyError:
            time.sleep(5)
            job.refresh()
    else:
        raise AssertionError("Failed to get result items")


@pytest.mark.vcr
def test_get_jobs(svc: Ades):
    jobs = svc.get_jobs()
    assert isinstance(jobs, PaginatedList)
    assert all(isinstance(job, Job) for job in jobs)


@pytest.mark.vcr
def test_get_job(svc: Ades):
    jobs = svc.get_jobs().get_limited()
    if len(jobs) > 0:
        first_id = jobs[0].id
        assert first_id is not None
        job = svc.get_job(first_id)
        assert isinstance(job, Job)
        assert job.id == first_id


@pytest.mark.vcr
def test_delete_job(svc: Ades):
    p = svc.get_process("convert-url")
    j = p.execute({})
    j.delete()
    assert j.status == AdesJobStatus.DISMISSED.value



import json
import logging
import os
import urllib.parse
from collections.abc import Callable
from typing import Any

import requests

from pyeodh import consts
from pyeodh.ades import Ades
from pyeodh.resource_catalog import CatalogService
from pyeodh.types import Headers, Params, RequestMethod
from pyeodh.utils import is_absolute_url
from pyeodh.workspace import Workspace

logger = logging.getLogger(__name__)


def _encode_json(data: dict) -> tuple[str, str]:
    return "application/json", json.dumps(data)


class Client:
    def __init__(
        self,
        base_url: str | None = None,
        username: str | None = None,
        token: str | None = None,
    ) -> None:
        """Initialize the client.

        Args:
            base_url (str, optional): Base URL of the EODH API. If not provided,
                the EODH_BASE_URL environment variable is used. Defaults to EODH
                production URL.
            username (str, optional): Username for authentication. Defaults to None.
            token (str, optional): Token for authentication. Defaults to None.
        """
        self.url_base = base_url or os.getenv("EODH_BASE_URL", consts.API_BASE_URL)

        if not is_absolute_url(self.url_base):
            raise ValueError("base_url parameter or EODH_BASE_URL environment variable must be an absolute URL")

        self.username = username
        self.token = token
        self.environment = self._get_environment(self.url_base)
        self._build_session()
        self.workspace = Workspace(self)

    def _get_environment(self, url: str) -> str:
        """Get the environment from the base URL.

        Returns:
            str: The environment
        """
        env = urllib.parse.urlparse(url).netloc.split(".")[0]
        logger.debug(f"Environment: {env}")
        if env == "eodatahub":  # prod environment doesn't have a subdomain
            return consts.Environment.PRODUCTION.value
        if env not in set(e.value for e in consts.Environment):
            raise ValueError(f"Invalid environment: {env}")
        return env

    def _build_session(
        self,
    ) -> None:
        self._session = requests.Session()
        if self.token is not None:
            self._session.headers.update({"Authorization": f"Bearer {self.token}"})

    def _request_raw(
        self,
        method: RequestMethod,
        url: str,
        headers: Headers | None = None,
        params: Params | None = None,
        data: Any | None = None,
        encode: Callable[[Any], tuple[str, Any]] = _encode_json,
    ) -> requests.Response:
        """Make a raw request.

        Args:
            method (RequestMethod): HTTP method to use
            url (str): Target URL
            headers (Optional[Headers], optional): Headers to send with the request.
                Defaults to None.
            params (Optional[Params], optional): Query parameters to send with the
                request. Defaults to None.
            data (Optional[Any], optional): Raw data to send with the request. Data is
                encoded by the `encode` function before sending. Defaults to None.
            encode (Callable[[Any], tuple[str, Any]], optional): Function to encode the
                data. Has to return a tuple of (content-type string, encoded data).
                Defaults to _encode_json.

        Returns:
            requests.Response: Response from the request
        """
        logger.debug(f"_request_raw received {locals()}")
        if not is_absolute_url(url):
            logger.debug(f"Received not absolute url: {url}")
            url = urllib.parse.urljoin(self.url_base, url)
            logger.debug(f"Created url from base: {url}")

        if ".org.uk/search" in url:
            url = url.replace(".org.uk/search", ".org.uk/api/catalogue/stac/search")

        headers = Headers() if headers is None else headers
        encoded_data = None
        if data is not None:
            headers["Content-Type"], encoded_data = encode(data)
        logger.debug(f"Making request: {method} {url}\nheaders: {headers}\nparams: {params}\nbody: {encoded_data}")
        response = self._session.request(
            method,
            url,
            headers=headers,
            params=params,
            data=encoded_data,
        )
        logger.debug(
            f"Received response {response.status_code}\nheaders: {response.headers}\ncontent: {response.text}"
        )
        # TODO consider moving this to _requst_json() and raise own exceptions
        # so that we can user _raw in e.g. delete methods where we expect a 409 and
        # want to recover
        response.raise_for_status()

        return response

    def _request_json(
        self,
        method: RequestMethod,
        url: str,
        headers: Headers | None = None,
        params: Params | None = None,
        data: Any | None = None,
        encode: Callable[[Any], tuple[str, Any]] = _encode_json,
    ) -> tuple[Headers, Any]:
        """Make a request and return the headers and deserialized JSON data. Input data
        can be anything, not just JSON, providing an `encode` function is supplied. Use
        when expecting JSON data in the response.

        Args:
            method (RequestMethod): HTTP method to use
            url (str): Target URL
            headers (Optional[Headers], optional): Headers to send with the request.
                Defaults to None.
            params (Optional[Params], optional): Query parameters to send with the
                request. Defaults to None.
            data (Optional[Any], optional): Raw data to send with the request. Data is
                encoded by the `encode` function before sending. Defaults to None.
            encode (Callable[[Any], tuple[str, Any]], optional): Function to encode the
                data. Has to return a tuple of (content-type string, encoded data).
                Defaults to _encode_json.

        Returns:
            tuple[Headers, Any]: Response headers and deserialized JSON data
        """
        response = self._request_raw(method, url, headers, params, data, encode)

        if not response.text:
            return response.headers, None

        return response.headers, json.loads(response.text)

    def get_catalog_service(self) -> CatalogService:
        """Initializes the resource catalog API client.

        Calls: GET /api/catalogue/stac

        Returns:
            CatalogService: Object representing the Resource catalog service.
        """
        headers, data = self._request_json("GET", "/api/catalogue/stac")
        return CatalogService(self, headers, data)

    def get_ades(self) -> Ades:
        """Initializes the workflow execution service (ADES) client.

        Calls: GET /ades/ogc-api

        Returns:
            Ades: Object representing the ADES.
        """
        if self.username is None or self.token is None:
            raise ValueError("Valid username and token required for accessing protected API endpoints.")
        data = {
            "links": [
                {
                    "rel": "self",
                    "href": (f"{self.url_base}/api/catalogue/stac/catalogs/user/catalogs/{self.username}"),
                }
            ]
        }
        return Ades(self, Headers(), data)



import pytest
from pytest_mock import MockerFixture

import pyeodh
from pyeodh.workspace import Workspace


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["Authorization"],
        "decode_compressed_response": True,
    }


@pytest.fixture
def workspace(api_token: str, username: str) -> Workspace:
    return pyeodh.Client(username=username, token=api_token, base_url="https://staging.eodatahub.org.uk").workspace


@pytest.mark.vcr
@pytest.mark.parametrize(
    "file, ws_file_path",
    [
        (b"test content", "test.txt"),
        ("tests/data/test.txt", "test.txt"),
    ],
)
def test_upload_file(workspace: Workspace, file: str | bytes, ws_file_path: str) -> None:
    workspace.upload_file(file=file, ws_file_path=ws_file_path)


def test_upload_nonexistent_file(workspace: Workspace) -> None:
    with pytest.raises(FileNotFoundError):
        workspace.upload_file(
            file="/path/to/nonexistent/file",
            ws_file_path="test.txt",
        )


def test_upload_file_mocked(mocker: MockerFixture, workspace: Workspace) -> None:
    # Test data
    test_content = b"test content"
    workspace_name = "test-workspace"
    ws_file_path = "test.txt"

    # Mock the _request_raw method
    mocker.patch.object(workspace._client, "_request_raw")
    spy = mocker.spy(workspace._client, "_request_raw")

    workspace.upload_file(file=test_content, ws_file_path=ws_file_path, workspace_name=workspace_name)
    assert spy.call_args.args[0] == "PUT"
    assert (
        spy.call_args.args[1] == f"https://{workspace_name}.staging.eodatahub-workspaces.org.uk/files/"
        f"workspaces-eodhp-staging/{ws_file_path}"
    )
    assert spy.call_args.kwargs["data"] == test_content



from typing import Any

import pytest

from pyeodh.eodh_object import is_optional
from pyeodh.utils import is_absolute_url, join_url, remove_null_items, s3_url


@pytest.mark.parametrize(
    "url, expected",
    [
        ("https://example.com", True),
        ("http://example.com", True),
        ("example.com", False),
        ("/path/to/resource", False),
        ("path/to/resource", False),
        ("", False),
    ],
)
def test_is_absolute_url(url: str, expected: bool) -> None:
    assert is_absolute_url(url) == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        (
            ("https://example.com", "path", "to", "resource"),
            "https://example.com/path/to/resource",
        ),
        (
            ("https://example.com/", "path", "to", "resource"),
            "https://example.com/path/to/resource",
        ),
        (("https://example.com", "/path", "to", "resource"), ValueError),
        (
            ("https://example.com", "path/with/slash/", "resource"),
            "https://example.com/path/with/slash/resource",
        ),
        (("", "path", "to", "resource"), "path/to/resource"),
        (("/path", "to", "resource"), "/path/to/resource"),
    ],
)
def test_join_url(args: tuple[str, ...], expected: str | type[ValueError]) -> None:
    if expected is ValueError:
        with pytest.raises(ValueError):
            join_url(*args)
    else:
        assert join_url(*args) == expected


@pytest.mark.parametrize(
    "input_dict, expected_dict",
    [
        ({"a": 1, "b": None, "c": 3}, {"a": 1, "c": 3}),
        ({"a": None, "b": None, "c": None}, {}),
        ({"a": 1, "b": 2, "c": 3}, {"a": 1, "b": 2, "c": 3}),
        ({"a": 0, "b": "", "c": False, "d": None}, {"a": 0, "b": "", "c": False}),
        ({}, {}),
    ],
)
def test_remove_null_items(input_dict: dict[str, Any], expected_dict: dict[str, Any]) -> None:
    assert remove_null_items(input_dict) == expected_dict


@pytest.mark.parametrize(
    "workspace_name, environment, path_to_file, expected",
    [
        # Basic case with staging environment
        (
            "test-workspace",
            "staging",
            "data/file.txt",
            "https://test-workspace.staging.eodatahub-workspaces.org.uk/files/workspaces-eodhp-staging/data/file.txt",
        ),
        # Path with leading slash
        (
            "test-workspace",
            "staging",
            "/data/file.txt",
            "https://test-workspace.staging.eodatahub-workspaces.org.uk/files/workspaces-eodhp-staging/data/file.txt",
        ),
        # Path with multiple leading slashes
        (
            "test-workspace",
            "staging",
            "///data/file.txt",
            "https://test-workspace.staging.eodatahub-workspaces.org.uk/files/workspaces-eodhp-staging/data/file.txt",
        ),
        # Test with test environment
        (
            "test-workspace",
            "test",
            "data/file.txt",
            "https://test-workspace.test.eodatahub-workspaces.org.uk/files/workspaces-eodhp-test/data/file.txt",
        ),
    ],
)
def test_s3_url(workspace_name: str, environment: str, path_to_file: str, expected: str) -> None:
    """Test s3_url function with various input combinations.

    Args:
        workspace_name: Name of the workspace to test
        path_to_file: Path to file within workspace to test
        expected: Expected S3 URL output
    """
    result = s3_url(workspace_name, environment, path_to_file)
    assert result == expected


@pytest.mark.parametrize(
    "value, type_, expected",
    [
        # Test with single type
        (None, str, True),
        ("hello", str, True),
        (123, str, False),
        # Test with tuple of types
        (None, (str, int), True),
        ("hello", (str, int), True),
        (123, (str, int), True),
        (1.23, (str, int), False),
        # Test with various Python types
        (None, dict, True),
        ({}, dict, True),
        ([], dict, False),
        # Test with complex types
        (None, list, True),
        ([], list, True),
        ({}, list, False),
    ],
)
def test_is_optional(value: Any, type_: type | tuple[type, ...], expected: bool) -> None:
    """Test is_optional function with various input combinations.

    Args:
        value: The value to test
        type_: The type or tuple of types to check against
        expected: Expected boolean result

    The function tests if a value is either None or matches the specified type(s).
    """
    assert is_optional(value, type_) == expected
